#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2016
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
Contains a slurm client which generates slurm scripts from daliuge graphs.
"""

import datetime
import sys
import os
import subprocess
import shutil
import tempfile
import string
import dlg.remote as remote
from dlg.runtime import __git_version__ as git_commit

from dlg.deploy.configs import ConfigFactory, init_tpl, dlg_exec_str
from dlg.deploy.configs import DEFAULT_MON_PORT, DEFAULT_MON_HOST
from dlg.deploy.deployment_utils import find_numislands, label_job_dur
from paramiko.ssh_exception import SSHException


class SlurmClient:
    """
    parameters we can control:

    1. user group / account name (Required)
    2. whether to submit a graph, and if so provide graph path
    3. # of nodes (of Drop Managers)
    4. how long to run
    5. whether to produce offline graph vis
    6. whether to attach proxy for remote monitoring, and if so provide
        DLG_MON_HOST
        DLG_MON_PORT
    7. Root directory of the Log files (Required)
    """

    def __init__(
        self,
        dlg_root=None,
        log_root=None,
        host=None,
        acc=None,
        physical_graph_template_file=None,  # filename of physical graph template
        logical_graph=None,
        job_dur=30,
        num_nodes=None,
        run_proxy=False,
        mon_host=DEFAULT_MON_HOST,
        mon_port=DEFAULT_MON_PORT,
        logv=1,
        facility=None,
        zerorun=False,
        max_threads=0,
        sleepncopy=False,
        num_islands=None,
        all_nics=False,
        check_with_session=False,
        submit=False,
        remote=True,
        pip_name=None,
        username=None,
        ssh_key="",
        config=None,
        slurm_template=None,
        suffix=None
    ):
        ## Here, we want to separate out the following
        ## Config derived from CONFIG Factory - we replace with ini file
        ## Config derived from CLI, intended for replacement in the SLURM job script
        ##    - We want to replace these directives with the SLURM template
        ## Config derived from CLI that is used in the final script call
        ## Any leftover config - we keep as normal

        if config:
            # Do the config from the config file
            self.host = config['login_node']
            self._acc = config['account'] # superceded by slurm_template if present
            self.dlg_root = config['dlg_root']
            self.modules = config['modules']
            self.venv = config['venv'] # superceded by slurm_template if present
            self.exec_prefix = config["exec_prefix"]
            self.username = config['user'] if 'user' in config else sys.exit(1)
            if not self.username:
                print("Username not configured in INI file, using local username...")
        else:
            # Setup SLURM environment variables using config
            config = ConfigFactory.create_config(facility=facility, user=username)
            self.host = config.getpar("host") if host is None else host
            self._acc = config.getpar("account") if (acc is None) else acc
            # self._user = config.getpar("user") if (username is None) else username

            # environment & sbatch
            self.dlg_root = config.getpar("dlg_root") if not dlg_root else dlg_root
            self.modules = config.getpar("modules")
            self.venv = config.getpar("venv")
            self.exec_prefix = config.getpar("exec_prefix")
            self.username = username
        # sbatch 
        if slurm_template:
            self._slurm_template = slurm_template
            self._num_nodes = 1 # placeholder
            self._job_dur = 1 # placeholder
        else:
            self._slurm_template = None
            if num_nodes is None:
                self._num_nodes = 1
            else:
                self._num_nodes = num_nodes
            self._job_dur = job_dur

        # self._log_root = (
        #     self._config.getpar("log_root") if (log_root is None) else log_root
        # )
        # 
        # start_dlg_cluster arguments
        self._logical_graph = logical_graph
        self._physical_graph_template_file = physical_graph_template_file
        self._visualise_graph = False
        self._run_proxy = run_proxy
        self._mon_host = mon_host
        self._mon_port = mon_port
        self._pip_name = pip_name
        self._logv = logv
        self._zerorun = zerorun
        self._max_threads = max_threads
        self._sleepncopy = sleepncopy
        if num_islands is None:
            self._num_islands = 1
        else:
            self._num_islands = num_islands
        self._all_nics = all_nics
        self._check_with_session = check_with_session
        self._submit = submit
        self._suffix = self.create_session_suffix(suffix)
        ni, nn, self._pip_name = find_numislands(self._physical_graph_template_file)
        if isinstance(ni, int) and ni >= self._num_islands:
            self._num_islands = ni
        if nn and nn >= self._num_nodes:
            self._num_nodes = nn

        # used for remote login/directory management.
        self._remote = remote
        self.ssh_key = ssh_key 
    
    def create_session_suffix(self, suffix=None):
        """
        Create a suffix to identify the session. If no suffix is specified, use the 
        current datetime setting. 

        :param: suffix, used to specify a non-datetime suffix. 
        :return: the final suffix 
        """
        if not suffix:
            return datetime.datetime.now().strftime("%Y-%m-%dT%H-%M-%S")
        else:
            return suffix

    def get_session_dirname(self):
        """
        (pipeline name_)[Nnum_of_daliuge_nodes]_[time_stamp]
        """
        graph_name = self._pip_name.split("_")[0]  # use only the part of the graph name
        graph_name = graph_name.rsplit(".pgt.graph")[0]
        return "{0}_{1}".format(graph_name, self._suffix)
    

    def apply_slurm_template(self, template_str, session_id, dlg_root):
        """
        Given a string from a template file, use a string.Template object to perform
        safe substution on the string and replace $VALUES with the correct value 
        specified. 
        """
        intermed_slurm = string.Template(template_str) 
        ims = intermed_slurm.safe_substitute(session_id=session_id, dlg_root=dlg_root)
        print("Creating job description")
        return ims + "\n\n" + dlg_exec_str

    def create_paramater_mapping(self, session_dir, physical_graph_file):
        """
        Map the runtime or configured parameters to the session environment and SLURM 
        script paramteres, in anticipation of using substition. 
        """ 
        pardict = {}
        pardict["SESSION_ID"] = os.path.split(session_dir)[-1]
        pardict["MODULES"] = self.modules
        pardict["DLG_ROOT"] = self.dlg_root
        pardict["EXEC_PREFIX"] = self.exec_prefix
        pardict["NUM_NODES"] = str(self._num_nodes)
        pardict["JOB_DURATION"] = label_job_dur(self._job_dur)

        pardict["VENV"] = self.venv
        pardict["PIP_NAME"] = self._pip_name
        pardict["ACCOUNT"] = self._acc
        pardict["PY_BIN"] = "python3" if pardict["VENV"] else sys.executable
        pardict["LOG_DIR"] = session_dir
        pardict["GRAPH_PAR"] = (
            '--logical-graph "{0}"'.format(self._logical_graph)
            if self._logical_graph
            else (
                '--physical-graph "{0}"'.format(physical_graph_file)
                if physical_graph_file
                else ""
            )
        )
        pardict["PROXY_PAR"] = (
            "--monitor_host %s --monitor_port %d" % (self._mon_host, self._mon_port)
            if self._run_proxy
            else ""
        )
        pardict["GRAPH_VIS_PAR"] = "--dump" if self._visualise_graph else ""
        pardict["LOGV_PAR"] = "--verbose-level %d" % self._logv
        pardict["ZERORUN_PAR"] = "--zerorun" if self._zerorun else ""
        pardict["MAXTHREADS_PAR"] = "--max-threads %d" % self._max_threads
        pardict["SNC_PAR"] = "--app 1" if self._sleepncopy else "--app 0"
        pardict["NUM_ISLANDS_PAR"] = "--num_islands %d" % self._num_islands
        pardict["ALL_NICS"] = "--all_nics" if self._all_nics else ""
        pardict["CHECK_WITH_SESSION"] = (
            "--check_with_session" if self._check_with_session else ""
        )
        return pardict

    def create_job_desc(self, physical_graph_file):
        """
        Creates the slurm script from a physical graph

        This uses string.Template to apply substitutions that are linked to the 
        parameters defined at runtime. These parameters map to $VALUEs in a pre-defined
        execution command that contains the necessary parameters to run DALiuGE through
        SLURM. 
        """

        session_dir = "{0}/workspace/{1}".format(
            self.dlg_root, self.get_session_dirname()
        )
        pardict = self.create_paramater_mapping(session_dir, physical_graph_file)

        if self._slurm_template:
            slurm_str = self.apply_slurm_template(self._slurm_template, 
                                                  pardict['SESSION_ID'],
                                                  pardict['DLG_ROOT'])
            return string.Template(slurm_str).safe_substitute(pardict)

        return init_tpl.safe_substitute(pardict)

    @property
    def session_dir(self):
        return "{0}/workspace/{1}".format(
            self.dlg_root, self.get_session_dirname()
        )

    def mk_session_dir(self, dlg_root: str = ""):
        """
        Create the session directory. If dlg_root is provided it is used,
        else env var DLG_ROOT is used.
        """

        if dlg_root:  # has always preference
            self.dlg_root = dlg_root
        if self._remote and not self.dlg_root:
            print("Deploying on a remote cluster requires specifying DLG_ROOT!")
            print("Unable to create session directory!")
            return ""
        elif not self._remote:
            # locally fallback to env var
            if os.environ["DLG_ROOT"]:
                dlg_root = os.environ["DLG_ROOT"]
            else:
                dlg_root = f"{os.environ['HOME']}.dlg"
        session_dir =  self.session_dir
        if not self._remote and not os.path.exists(session_dir):
            os.makedirs(session_dir)
        if self._remote:
            command = f"mkdir -p {session_dir}"
            print(
                f"Creating remote session directory on {self.username}@{self.host}: {command}"
            )
            try:
                remote.execRemote(
                    self.host, command, username=self.username, pkeyPath=self.ssh_key
                )
            except (TypeError, SSHException) as e:
                print(
                    f"ERROR: Unable to create {session_dir} on {self.username}@{self.host}, {str(e)}"
                )
                return None

        return session_dir

    def submit_job(self):
        """
        Submits the slurm script to the requested facility

        :returns: jobId, the id of the SLURM job create on the facility. 
                  None if a remote directory could not be created or if an error occurs
                  during connection. 
        """
        jobId = None
        session_dir = self.mk_session_dir()
        if not session_dir:
            print("No session_dir created.")
            return jobId

        physical_graph_file_name = "{0}/{1}".format(session_dir, self._pip_name)
        if self._physical_graph_template_file:
            if self._remote:
                print(f"Copying PGT to: {physical_graph_file_name}")
                remote.copyTo(
                    self.host,
                    self._physical_graph_template_file,
                    physical_graph_file_name,
                    username=self.username,
                    pkeyPath=self.ssh_key,
                )
            else:
                shutil.copyfile(
                    self._physical_graph_template_file, physical_graph_file_name
                )

        job_file_name = "{0}/jobsub.sh".format(session_dir)
        job_desc = self.create_job_desc(physical_graph_file_name)

        if self._remote:
            print(f"Creating SLURM script remotely: {job_file_name}")
            tjob = tempfile.mktemp()
            with open(tjob, "w+t") as t:
                t.write(job_desc)
            remote.copyTo(
                self.host,
                tjob,
                job_file_name,
                username=self.username,
                pkeyPath=self.ssh_key,
            )
            os.remove(tjob)
        else:
            with open(job_file_name, "w") as job_file:
                job_file.write(job_desc)
            with open(os.path.join(session_dir, "git_commit.txt"), "w") as git_file:
                git_file.write(git_commit)
        if self._submit:
            if not self._remote:
                os.chdir(session_dir)  # so that slurm logs will be dumped here
                print(subprocess.check_output(["sbatch", job_file_name]))
            else:
                command = f"cd {session_dir} && sbatch --parsable {job_file_name}"
                print(f"Submitting sbatch job: {command}")
                stdout, stderr, exitStatus = remote.execRemote(
                    self.host, command, username=self.username, pkeyPath=self.ssh_key
                )
                if exitStatus != 0:
                    print(
                        f"Job submission unsuccessful: {exitStatus}, {stderr.decode()}"
                    )
                else:
                    jobId = stdout.decode()
                    print(f"Job with ID {jobId} submitted successfully.")
        else:
            print(f"Created job submission script {job_file_name}")
        return jobId



# THIS FILE IS GENERATED BY SETUP.PY
# DO NOT MODIFY BY HAND
version = '5.5.0'
git_version = 'ce1d8837be043f3e5f66b96cfb0f59e9425ec598'
full_version = '5.5.0'
is_release = True

if not is_release:
    version = full_version


#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2025
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
import sys
import optparse
import shutil
import textwrap
import typing

from enum import Enum, auto
from dataclasses import dataclass
from importlib.resources import files
from pathlib import Path

import dlg.deploy.configs as default_configs
from dlg.deploy.configs import ConfigFactory

USER_CONFIG_DIR = Path.home() / ".config"
DLG_CONFIG_DIR = USER_CONFIG_DIR / "dlg"

class ConfigType(Enum):
    ENV = "ini"
    SLURM = "slurm"


class ConfigDirState(Enum):
    NEW = auto()
    EXISTS = auto()
    NOT_CREATED = auto()


@dataclass
class ConfigManager:
    """
    Utility class to manager job submission configuration files
    """

    facilities: typing.List[str]
    prompt_setup = True

    def setup_user(self):
        """
        Called when passing the "--setup" command of the "dlg remote" option.
        """
        dir_state = self.create_dlg_config_directory()
        if dir_state == ConfigDirState.NEW:
            self.copy_defaults_to_user_config_directory()
            self.print_available_config()
        elif dir_state == ConfigDirState.EXISTS:
            print(f"{DLG_CONFIG_DIR} already exists.")
        else:
            print(f"{DLG_CONFIG_DIR} was not created.")

    def prompt_user(self):
        """
        Create prompt and validate it against expected output
        """

        while True:
            ui = input("Do you want to create a $HOME/.config/dlg directory"
                       " to store your custom configuration files and scripts (y/n)? ")
            if ui.lower() == "y" or ui.lower() == "n":
                return ui.lower() == "y"
            else:
                print("Please selected from the options (Y/N)")

    def create_dlg_config_directory(self):
        """
        Establish the user $HOME/.config/dlg/ directory
        """
        if DLG_CONFIG_DIR.exists():
            return ConfigDirState.EXISTS
        else:
            if self.prompt_user():
                try:
                    Path.mkdir(DLG_CONFIG_DIR)
                    return ConfigDirState.NEW
                except Exception as e:
                    raise (e)
            else:
                return ConfigDirState.NOT_CREATED

    def copy_defaults_to_user_config_directory(self):
        """
        Move configuration files from dlg/deploy/configs to DLG_CONFIG_DIR
        """
        cfg_path = files(default_configs)

        configs = [p for p in cfg_path.iterdir() if
                   (".ini" in p.name) or (".slurm" in p.name)]

        if DLG_CONFIG_DIR.exists():
            for cfg in configs:
                if cfg.exists():
                    shutil.copy(cfg, DLG_CONFIG_DIR)
                else:
                    print("Unable to copy %s, does not exist", cfg)
        else:
            print("Unable to copy to %s, does is not available", DLG_CONFIG_DIR)

    def get_user_configs(self) -> typing.Dict:
        """
        Returns dictionary of filetypes to split out?
        """
        configs = {}
        if not DLG_CONFIG_DIR.exists():
            return configs
        filetypes = ["ini", "slurm"]

        for ftype in filetypes:
            configs[ftype] = [cfg for cfg in DLG_CONFIG_DIR.iterdir() if
                ftype in cfg.name]
        return configs

    def print_available_config(self):
        """
        Present the options available for deployment
        """
        print("User Configs (~/.config/dlg)")
        print("----------------------------")
        user_configs = self.get_user_configs()
        if user_configs:
            print("Environments (--config_file):")
            for config in user_configs["ini"]:
                print(textwrap.indent(config.name, "\t"))
            print("Slurm scripts (--slurm_template):")
            for config in user_configs["slurm"]:
                print(textwrap.indent(config.name, "\t"))
        else:
            print(textwrap.indent("N/a: User-specific directory is not setup.", "\t"))

        print("\nDALiuGE Defaults (-f/--facility):")
        print("-----------------------------------")
        for f in self.facilities:
            print(textwrap.indent(f, "\t"))

    def load_user_config(self, config_type: ConfigType, config_choice: str) -> Path:
        """
        Resolve the config type and choice to a config file and return full path.

        Path resolution occurs in the following hierarchy:
        - If config choice is absolute path, return the absolute path
        - If config choice is not an absolute path, but the path exists (i.e. can be
        found) return the path
        - If the config choice is relative, and doesn't exist, check if it exists in
          DLG_CONFIG_DIR
            - If it does exist, construct the path
            - If it doesn't exist, return None
        """
        if not DLG_CONFIG_DIR.exists() and self.prompt_setup:
            print("NOTE: No user configs exists; consider running 'dlg config --setup'.")
            self.prompt_setup = False

        choice_path = Path(config_choice)
        if choice_path.is_absolute() and choice_path.exists():
            return choice_path
        elif choice_path.exists():
            return choice_path.absolute()
        else:
            user_configs = self.get_user_configs()

            options = user_configs.get(config_type.value, [])
            for o in options:
                if config_choice == o.name:
                    return o
            return None

def run(_, args):

    cfg_manager = ConfigManager(ConfigFactory.available())

    parser = optparse.OptionParser()
    parser.add_option(
        "--setup",
        dest="setup",
        action="store_true",
        help="Setup local '$HOME/.config/dlg' directory to store custom environment config and slurm scripts",
        default=False
    )
    parser.add_option(
        "-l", "--list",
        dest="list",
        action="store_true",
        help="List the available configuration for DALiuGE deployment."
    )
    (opts, _) = parser.parse_args(sys.argv)
    if opts.setup:
        cfg_manager.setup_user()
        sys.exit(0)
    elif opts.list:
        print(f"Available facilities:\n")
        cfg_manager.print_available_config()
        sys.exit(0)
    else:
        parser.print_help()






#
#    ICRAR - International Centre for Radio Astronomy Research
#    (c) UWA - The University of Western Australia, 2015
#    Copyright by UWA (in the framework of the ICRAR)
#    All rights reserved
#
#    This library is free software; you can redistribute it and/or
#    modify it under the terms of the GNU Lesser General Public
#    License as published by the Free Software Foundation; either
#    version 2.1 of the License, or (at your option) any later version.
#
#    This library is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with this library; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#    MA 02111-1307  USA
#
"""
A module containing utility code for running remote commands over SSH.
"""

import logging
import os
import sys
import time

from paramiko.client import SSHClient, AutoAddPolicy
from paramiko.rsakey import RSAKey
import scp
from typing import Union

logger = logging.getLogger(f"dlg.{__name__}")


def execRemoteWithClient(client, command, timeout=None, bufsize=-1):
    """
    Executes `command` using the given SSH `client`.
    """
    chan = client.get_transport().open_session()
    chan.settimeout(timeout)
    chan.exec_command('/bin/bash -l -c "%s"' % (command.replace('"', '"')))
    # Otherwise do something like this:
    # chan.get_pty(width=80, height=24)
    # chan.invoke_shell()
    # chan.sendall(command + "\n")

    # Wait until the command has finished execution and return the full contents
    # of the stdout and stderr
    while True:
        if chan.exit_status_ready():
            break
        try:
            time.sleep(0.01)
        except KeyboardInterrupt:
            chan.send("\x03")

    exitStatus = chan.recv_exit_status()
    with chan.makefile("r", bufsize) as f:
        stdout = f.read()
    with chan.makefile_stderr("r", bufsize) as f:
        stderr = f.read()
    return stdout, stderr, exitStatus


def execRemote(
    host: str,
    command: str,
    username: str = None,
    pkeyPath: str = None,
    timeout: int = None,
    bufsize: int = -1,
):
    """
    Executes `command` on `host`. If `username` is provided, the command will
    be run as `username`; otherwise the current user will be used to run the
    command. Returns the tuple (stdout, stderr, exitStatus)

    Parameters:
    -----------
    host: str, Name of the host to connect to
    command: str, Command to execute
    username: str, Username to be used for connection, if None current user is used
    pkeyPath: str, Path to private key file for connection
    timeout: int, Timeout in seconds, None is default
    bufsize: int, Size of buffer for the return, unlimited by default

    Returns:
    --------
    tuple, (stdout, stderr, exitStatus)
    """
    with createClient(host, username, pkeyPath=pkeyPath) as client:
        return execRemoteWithClient(client, command, timeout, bufsize)


def createClient(host, username=None, pkeyPath=None):
    """
    Creates an SSH client object that can be used to perform SSH-related
    operations
    """
    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy())
    try:
        client.connect(host, username=username, key_filename=pkeyPath)
    except FileNotFoundError:
        logger.warning(
            "File '%s' was not found, cannot create remote connection", pkeyPath
        )
        sys.exit(1)
    return client


def __scpProgress(filename, size, sent):
    if size == sent:
        logger.debug("Finished scp-ing %s (%d bytes)", filename, sent)


def copyFrom(
    host,
    remotePath,
    localPath=".",
    recursive=False,
    username=None,
    pkeyPath=None,
    timeout=None,
):
    """
    Copies the files located at `host`:`remotePath` to `localPath` connecting
    to `host` as `username`. A `recursive` flag can be specified, as well as a
    private key (via `pkeyPath`) to be used when creating the connection.
    """
    if timeout is None:
        timeout = 5.0
    client = createClient(host, username=username, pkeyPath=pkeyPath)
    with scp.SCPClient(
        client.get_transport(), progress=__scpProgress, socket_timeout=timeout
    ) as scpClient:
        scpClient.get(remote_path=remotePath, local_path=localPath, recursive=recursive)
    client.close()


def copyTo(
    host,
    localFiles,
    remotePath=".",
    recursive=False,
    username=None,
    pkeyPath=None,
    timeout=None,
):
    """
    Copies the files located at `localPath` to `host`:`remotePath` connecting
    to `host` as `username`. A `recursive` flag can be specified, as well as a
    private key (via `pkeyPath`) to be used when creating the connection.
    """
    if timeout is None:
        timeout = 5.0
    client = createClient(host, username=username, pkeyPath=pkeyPath)
    with scp.SCPClient(
        client.get_transport(), progress=__scpProgress, socket_timeout=timeout
    ) as scpClient:
        scpClient.put(localFiles, remote_path=remotePath, recursive=recursive)
    client.close()
